#include "Matrix.h"
#include <iostream>
#include <cmath>
#include <cstring>

static void makeIdentity(float* mat)
{
    mat[0] = 1;    mat[4] = 0;    mat[8]  = 0;    mat[12] = 0;
    mat[1] = 0;    mat[5] = 1;    mat[9]  = 0;    mat[13] = 0;
    mat[2] = 0;    mat[6] = 0;    mat[10] = 1;    mat[14] = 0;
    mat[3] = 0;    mat[7] = 0;    mat[11] = 0;    mat[15] = 1;
}

Matrix::Matrix()
{
    setIdentity();
}

Matrix::Matrix(const Vector &right, const Vector &up, const Vector &forward)
{
    mat[0] = right.x();    mat[4] = right.y();    mat[8]  = right.z();    mat[12] = 0;
    mat[1] = up.x();       mat[5] = up.y();       mat[9]  = up.z();       mat[13] = 0;
    mat[2] = forward.x();  mat[6] = forward.y();  mat[10] = forward.z();  mat[14] = 0;
    mat[3] = 0;            mat[7] = 0;            mat[11] = 0;            mat[15] = 1;
}

Matrix::Matrix(GLdouble* otherMat)
{
    for (int i=0; i<16; i++)
    {
        mat[i] = otherMat[i];
    }
//    std::memcpy(&mat, mat, sizeof(float)*16);
}

Matrix::~Matrix()
{
}

void Matrix::setIdentity()
{
    makeIdentity(mat);
}

void Matrix::rotate(float degrees, Axis axis)
{
    float tempMat[16];
    float s = std::sin(degrees * M_PI / 180.0);
    float c = std::cos(degrees * M_PI / 180.0);

    switch(axis)
    {
    case X:
        tempMat[0] = 1; tempMat[4] = 0; tempMat[8]  = 0;  tempMat[12] = 0;
        tempMat[1] = 0; tempMat[5] = c; tempMat[9]  = -s; tempMat[13] = 0;
        tempMat[2] = 0; tempMat[6] = s; tempMat[10] = c;  tempMat[14] = 0;
        tempMat[3] = 0; tempMat[7] = 0; tempMat[11] = 0;  tempMat[15] = 1;
        break;
    case Y:
        tempMat[0] = c;  tempMat[4] = 0; tempMat[8]  = s; tempMat[12] = 0;
        tempMat[1] = 0;  tempMat[5] = 1; tempMat[9]  = 0; tempMat[13] = 0;
        tempMat[2] = -s; tempMat[6] = 0; tempMat[10] = c; tempMat[14] = 0;
        tempMat[3] = 0;  tempMat[7] = 0; tempMat[11] = 0; tempMat[15] = 1;
        break;
    case Z:
        tempMat[0] = c; tempMat[4] = -s; tempMat[8] = 0; tempMat[12] = 0;
        tempMat[1] = s; tempMat[5] = c; tempMat[9]  = 0; tempMat[13] = 0;
        tempMat[2] = 0; tempMat[6] = 0; tempMat[10] = 1; tempMat[14] = 0;
        tempMat[3] = 0; tempMat[7] = 0; tempMat[11] = 0; tempMat[15] = 1;
        break;
    }
    matrixMult(tempMat);
}

void Matrix::rotate(float degrees, const Vector &v)
{
    float matrix[16];
    // according to http://www.opengl.org/sdk/docs/man/xhtml/glRotate.xml
    GLfloat s = std::sin(degrees * M_PI / 180.0);
    GLfloat c = std::cos(degrees * M_PI / 180.0);
    GLfloat x = v.x();
    GLfloat y = v.y();
    GLfloat z = v.z();

    matrix[0] = x*x*(1-c)+c;
    matrix[1] = x*y*(1-c)+z*s;
    matrix[2] = x*z*(1-c)-y*s;
    matrix[3] = 0;
    matrix[4] = y*x*(1-c)-z*s;
    matrix[5] = y*y*(1-c)+c;
    matrix[6] = y*z*(1-c)+x*s;
    matrix[7] = 0;
    matrix[8] = z*x*(1-c)+y*s;
    matrix[9] = z*y*(1-c)-x*s;
    matrix[10] = z*z*(1-c)+c;
    matrix[11] = 0;
    matrix[12] = 0;
    matrix[13] = 0;
    matrix[14] = 0;
    matrix[15] = 1;

    matrixMult(matrix);
}

void Matrix::translate(const Vector& v)
{
    float tempMat[16];
    makeIdentity(tempMat);

    tempMat[12] = v.x();
    tempMat[13] = v.y();
    tempMat[14] = v.z();

    matrixMult(tempMat);
}

void Matrix::scale(const Vector& v)
{
    float tempMat[16];
    makeIdentity(tempMat);

    tempMat[0] = v.x();
    tempMat[5] = v.y();
    tempMat[10] = v.z();

    matrixMult(tempMat);
}

void Matrix::matrixMult(float* otherMat)
{
    float tempMat[16];
    tempMat[0] = mat[0] * otherMat[0] + mat[4] * otherMat[1] + mat[8]  * otherMat[2] + mat[12] * otherMat[3];
    tempMat[1] = mat[1] * otherMat[0] + mat[5] * otherMat[1] + mat[9]  * otherMat[2] + mat[13] * otherMat[3];
    tempMat[2] = mat[2] * otherMat[0] + mat[6] * otherMat[1] + mat[10] * otherMat[2] + mat[14] * otherMat[3];
    tempMat[3] = mat[3] * otherMat[0] + mat[7] * otherMat[1] + mat[11] * otherMat[2] + mat[15] * otherMat[3];

    tempMat[4] = mat[0] * otherMat[4] + mat[4] * otherMat[5] + mat[8]  * otherMat[6] + mat[12] * otherMat[7];
    tempMat[5] = mat[1] * otherMat[4] + mat[5] * otherMat[5] + mat[9]  * otherMat[6] + mat[13] * otherMat[7];
    tempMat[6] = mat[2] * otherMat[4] + mat[6] * otherMat[5] + mat[10] * otherMat[6] + mat[14] * otherMat[7];
    tempMat[7] = mat[3] * otherMat[4] + mat[7] * otherMat[5] + mat[11] * otherMat[6] + mat[15] * otherMat[7];

    tempMat[8]  = mat[0] * otherMat[8] + mat[4] * otherMat[9] + mat[8]  * otherMat[10] + mat[12] * otherMat[11];
    tempMat[9]  = mat[1] * otherMat[8] + mat[5] * otherMat[9] + mat[9]  * otherMat[10] + mat[13] * otherMat[11];
    tempMat[10] = mat[2] * otherMat[8] + mat[6] * otherMat[9] + mat[10] * otherMat[10] + mat[14] * otherMat[11];
    tempMat[11] = mat[3] * otherMat[8] + mat[7] * otherMat[9] + mat[11] * otherMat[10] + mat[15] * otherMat[11];

    tempMat[12] = mat[0] * otherMat[12] + mat[4] * otherMat[13] + mat[8]  * otherMat[14] + mat[12] * otherMat[15];
    tempMat[13] = mat[1] * otherMat[12] + mat[5] * otherMat[13] + mat[9]  * otherMat[14] + mat[13] * otherMat[15];
    tempMat[14] = mat[2] * otherMat[12] + mat[6] * otherMat[13] + mat[10] * otherMat[14] + mat[14] * otherMat[15];
    tempMat[15] = mat[3] * otherMat[12] + mat[7] * otherMat[13] + mat[11] * otherMat[14] + mat[15] * otherMat[15];

    std::memcpy(&mat, &tempMat, sizeof(float)*16);
}

void Matrix::mult(Matrix& other)
{
    matrixMult(other.mat);
}

Vector Matrix::operator* (const Vector& v) const
{
    return Vector(mat[0] * v.x() + mat[4] * v.y() + mat[8]  * v.z() + mat[12] * v.w(),
                  mat[1] * v.x() + mat[5] * v.y() + mat[9]  * v.z() + mat[13] * v.w(),
                  mat[2] * v.x() + mat[6] * v.y() + mat[10] * v.z() + mat[14] * v.w(),
                  mat[3] * v.x() + mat[7] * v.y() + mat[11] * v.z() + mat[15] * v.w());
}

Vector Matrix::right() const
{
    return Vector(mat[0], mat[4], mat[8]);
}

Vector Matrix::up() const
{
    return Vector(mat[1], mat[5], mat[9]);
}

Vector Matrix::forward() const
{
    return Vector(mat[2], mat[6], mat[10]);
}

Vector Matrix::translation() const
{
    return Vector(mat[12], mat[13], mat[14]);
}

void Matrix::debugPrint() const
{
    for (int i=0; i<4; i++)
    {
        for (int j=0; j<4; j++)
        {
            std::cout << mat[j*4+i] << ", ";
        }
        std::cout << std::endl;
    }
}

float* Matrix::get()
{
    return mat;
}


Matrix Matrix::fromViewDir(const Vector &forward)
{
    Vector out = -forward;
    Vector right = (Vector(0,1,0)).cross(out);
    if (right.length2() < 0.0001)
        right = (Vector(1,0,0)).cross(out);

    Vector up = out.cross(right);

    return Matrix(right.normalized(), up.normalized(), out.normalized());
}

static float clamp(float angle, float min, float max)
{
    if (angle < min)
        return max + angle;
    if (angle > max)
        return angle - max;
    else
        return angle;
}

//see http://www.flipcode.com/documents/matrfaq.html#Q37
void Matrix::euler(float* x, float* y, float* z) const
{
    float angle_x, angle_y, angle_z;
    float C, D;
    float RADIANS = 180.0/M_PI; //RadToDeg
    angle_y = D = -std::asin( mat[2]);        /* Calculate Y-axis angle */
    C        =  std::cos( angle_y );
    angle_y    *= RADIANS;

    if ( std::fabs( C ) > 0.0005 )        /* Gimbal lock? */
    {
        float tr_x      =  mat[10] / C;           /* No, so get X-axis angle */
        float tr_y      = -mat[6]  / C;

        angle_x  = std::atan2( tr_y, tr_x ) * RADIANS;

        tr_x      =  mat[0] / C;            /* Get Z-axis angle */
        tr_y      = -mat[1] / C;

        angle_z  = std::atan2( tr_y, tr_x ) * RADIANS;
    }
    else                                 /* Gimball lock has occurred */
    {
        angle_x  = 0;                      /* Set X-axis angle to zero */

        float tr_x      = mat[5];                 /* And calculate Z-axis angle */
        float tr_y      = mat[4];

        angle_z  = atan2( tr_y, tr_x ) * RADIANS;
    }
    std::cout << angle_x << " " << angle_y << " " << angle_z <<std::endl;
    angle_x = clamp( angle_x, 0, 360 );  /* Clamp all angles to range */
    angle_y = clamp( angle_y, 0, 360 );
    angle_z = clamp( angle_z, 0, 360 );
    *x = 360 - angle_x;
    *y = angle_y;
    *z = 360 - angle_z;

}
